/*
 * Copyright (C) 2013 Fairphone Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.fairphone.updater.data;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Locale;
import java.util.NoSuchElementException;
import java.util.Scanner;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

import android.content.Context;
import android.content.res.Resources;
import android.os.Environment;
import android.util.Log;

import com.fairphone.updater.R;

public class VersionParserHelper
{

    private static final String TAG = VersionParserHelper.class.getSimpleName();

    private static final String CURRENT_VERSION_NUMBER = "fairphone.ota.version.number";
    private static final String CURRENT_VERSION_NAME = "fairphone.ota.version.name";
    private static final String CURRENT_VERSION_BUILD_NUMBER = "fairphone.ota.build_number";
    private static final String CURRENT_ANDROID_VERSION = "fairphone.ota.android_version";
    private static final String CURRENT_VERSION_IMAGE_TYPE = "fairphone.ota.image_type";
    private static final String CURRENT_VERSION_BUILD_DATE = "ro.build.date.utc";

    public static Version getDeviceVersion(Context context)
    {

        Version version = new Version();

        try
        {
            version.setNumber(Integer.valueOf(getSystemData(context, CURRENT_VERSION_NUMBER)));
        } catch (NumberFormatException e)
        {
            version.setNumber(context.getResources().getInteger(R.integer.defaultVersionNumber));
        }
        version.setName(getSystemData(context, CURRENT_VERSION_NAME));
        version.setBuildNumber(getSystemData(context, CURRENT_VERSION_BUILD_NUMBER));
        version.setAndroidVersion(getSystemData(context, CURRENT_ANDROID_VERSION));
        version.setImageType(getSystemData(context, CURRENT_VERSION_IMAGE_TYPE));
        version.setReleaseDate(getSystemData(context, CURRENT_VERSION_BUILD_DATE));

        Version versionData = UpdaterData.getInstance().getVersion(version.getImageType(), version.getNumber());
        version.setThumbnailLink(versionData != null ? versionData.getThumbnailLink() : "");
        version.setReleaseNotes(Locale.getDefault().getLanguage(), versionData != null ? versionData.getReleaseNotes(Locale.getDefault().getLanguage()) : "");

        return version;
    }

    public static String getSystemData(Context context, String property)
    {

        if (property.equals(CURRENT_VERSION_NUMBER))
        {
            return getprop(CURRENT_VERSION_NUMBER, String.valueOf(context.getResources().getInteger(R.integer.defaultVersionNumber)));
        }
        else if (property.equals(CURRENT_VERSION_NAME))
        {
            return getprop(CURRENT_VERSION_NAME, context.getResources().getString(R.string.defaultVersionName));
        }
        else if (property.equals(CURRENT_ANDROID_VERSION))
        {
            return getprop(CURRENT_ANDROID_VERSION, context.getResources().getString(R.string.defaultAndroidVersionNumber));
        }
        else if (property.equals(CURRENT_VERSION_BUILD_NUMBER))
        {
            return getprop(CURRENT_VERSION_BUILD_NUMBER, context.getResources().getString(R.string.defaultBuildNumber));
        }
        else if (property.equals(CURRENT_VERSION_IMAGE_TYPE))
        {
            return getprop(CURRENT_VERSION_IMAGE_TYPE, context.getResources().getString(R.string.defaultImageType));
        }
        else if (property.equals(CURRENT_VERSION_BUILD_DATE))
        {
            return getprop(CURRENT_VERSION_BUILD_DATE, context.getResources().getString(R.string.defaultBuildDate));
        }

        return null;
    }

    public static Version getLatestVersion(Context context)
    {

        Version latest = null;
        Resources resources = context.getResources();
        String filePath =
                Environment.getExternalStorageDirectory() + resources.getString(R.string.updaterFolder) + resources.getString(R.string.configFilename);
        File file = new File(filePath + resources.getString(R.string.config_xml));

        if (file.exists())
        {
            try
            {
                latest = parseLatestXML(context, file);
            } catch (XmlPullParserException e)
            {
                Log.e(TAG, "Could not start the XML parser", e);
            } catch (IOException e)
            {
                Log.e(TAG, "Invalid data in File", e);
                // remove the files
                removeConfigFiles(context);
            }
        }

        removeZipContents(context);

        return latest;
    }

    // @formatter:off
    public enum XML_LEVEL_TAGS
    {
        NONE, AOSP, FAIRPHONE, STORES
    }

    public enum XML_TAGS
    {
        RELEASES, AOSP, FAIRPHONE, VERSION, NAME, BUILD_NUMBER, ANDROID_VERSION, RELEASE_NOTES, RELEASE_DATE, MD5SUM, THUMBNAIL_LINK, UPDATE_LINK, ERASE_DATA_WARNING, DEPENDENCIES, STORES, STORE, SHOW_DISCLAIMER;
    }

    // @formatter:on

    private static Version parseLatestXML(Context context, File latestFile) throws XmlPullParserException, IOException
    {

        Version version = null;
        Store store = null;

        UpdaterData update = UpdaterData.getInstance();
        update.resetUpdaterData();

        XmlPullParserFactory factory = XmlPullParserFactory.newInstance();
        factory.setNamespaceAware(true);

        FileInputStream fis = new FileInputStream(latestFile);

        XmlPullParser xpp = factory.newPullParser();

        xpp.setInput(new InputStreamReader(fis));

        int eventType = xpp.getEventType();

        XML_LEVEL_TAGS currentTag = XML_LEVEL_TAGS.NONE;

        while (eventType != XmlPullParser.END_DOCUMENT)
        {
            String tagName = null;
            switch (eventType)
            {
                case XmlPullParser.START_DOCUMENT:
                    break;
                case XmlPullParser.START_TAG:
                    tagName = xpp.getName();
                    currentTag = getCurrentXmlElement(tagName, currentTag);
                    switch (currentTag)
                    {
                        case AOSP:
                            if (tagName.equalsIgnoreCase(XML_TAGS.AOSP.name()))
                            {
                                update.setLatestAOSPVersionNumber(xpp.getAttributeValue(0));
                            }
                            version = readVersion(version, xpp, tagName);
                            version.setImageType(Version.IMAGE_TYPE_AOSP);
                            break;
                        case FAIRPHONE:
                            if (tagName.equalsIgnoreCase(XML_TAGS.FAIRPHONE.name()))
                            {
                                update.setLatestFairphoneVersionNumber(xpp.getAttributeValue(0));
                            }
                            version = readVersion(version, xpp, tagName);
                            version.setImageType(Version.IMAGE_TYPE_FAIRPHONE);
                            break;
                        case STORES:
                            store = readStore(store, xpp, tagName);
                            break;
                        default:
                            break;
                    }
                    break;
                case XmlPullParser.END_TAG:
                    tagName = xpp.getName();
                    currentTag = getCurrentXmlElement(tagName, currentTag);
                    switch (currentTag)
                    {
                        case AOSP:
                            if (tagName.equalsIgnoreCase(XML_TAGS.VERSION.name()))
                            {
                                update.addAOSPVersion(version);
                                version = null;
                            }
                            break;
                        case FAIRPHONE:
                            if (tagName.equalsIgnoreCase(XML_TAGS.VERSION.name()))
                            {
                                update.addFairphoneVersion(version);
                                version = null;
                            }
                            break;
                        case STORES:
                            if (tagName.equalsIgnoreCase(XML_TAGS.STORE.name()))
                            {
                                update.addAppStore(store);
                                store = null;
                            }
                            break;
                        default:
                            break;
                    }
                    break;
            }

            eventType = xpp.next();
        }
        fis.close();

        removeZipContents(context);
        return update.getLatestVersion(getSystemData(context, CURRENT_VERSION_IMAGE_TYPE));
    }

    public static Version readVersion(Version version, XmlPullParser xpp, String tagName) throws XmlPullParserException, IOException
    {

        if (version == null)
        {
            version = new Version();
        }

        readDownloadableItem(version, xpp, tagName);

        if (tagName.equalsIgnoreCase(XML_TAGS.VERSION.name()))
        {
            version.setNumber(xpp.getAttributeValue(0));
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.ANDROID_VERSION.name()))
        {
            version.setAndroidVersion(xpp.nextText());
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.ERASE_DATA_WARNING.name()))
        {
            version.setEraseAllPartitionWarning(true);
        }

        return version;
    }

    public static Store readStore(Store store, XmlPullParser xpp, String tagName) throws XmlPullParserException, IOException
    {

        if (store == null)
        {
            store = new Store();
        }

        readDownloadableItem(store, xpp, tagName);

        if (tagName.equalsIgnoreCase(XML_TAGS.STORE.name()))
        {
            store.setNumber(xpp.getAttributeValue(0));
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.SHOW_DISCLAIMER.name()))
        {
            store.setShowDisclaimer(true);
        }

        return store;
    }

    private static void readDownloadableItem(DownloadableItem item, XmlPullParser xpp, String tagName) throws XmlPullParserException, IOException
    {
        if (tagName.equalsIgnoreCase(XML_TAGS.NAME.name()))
        {
            item.setName(xpp.nextText());
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.BUILD_NUMBER.name()))
        {
            item.setBuildNumber(xpp.nextText());
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.DEPENDENCIES.name()))
        {
            item.setVersionDependencies(xpp.nextText());
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.RELEASE_NOTES.name()))
        {
            item.setReleaseNotes(Version.DEFAULT_NOTES_LANG, xpp.nextText());
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.RELEASE_NOTES.name() + "_" + Locale.getDefault().getLanguage()))
        {
            item.setReleaseNotes(Locale.getDefault().getLanguage(), xpp.nextText());
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.RELEASE_DATE.name()))
        {
            item.setReleaseDate(xpp.nextText());
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.MD5SUM.name()))
        {
            item.setMd5Sum(xpp.nextText());
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.THUMBNAIL_LINK.name()))
        {
            item.setThumbnailLink(xpp.nextText());
        }
        else if (tagName.equalsIgnoreCase(XML_TAGS.UPDATE_LINK.name()))
        {
            item.setDownloadLink(xpp.nextText());
        }
    }

    private static XML_LEVEL_TAGS getCurrentXmlElement(String tagName, XML_LEVEL_TAGS current)
    {
        XML_LEVEL_TAGS retval = current;

        if (tagName.equalsIgnoreCase(XML_LEVEL_TAGS.AOSP.name()))
        {
            retval = XML_LEVEL_TAGS.AOSP;
        }
        else if (tagName.equalsIgnoreCase(XML_LEVEL_TAGS.FAIRPHONE.name()))
        {
            retval = XML_LEVEL_TAGS.FAIRPHONE;
        }
        else if (tagName.equalsIgnoreCase(XML_LEVEL_TAGS.STORES.name()))
        {
            retval = XML_LEVEL_TAGS.STORES;
        }
        return retval;
    }

    private static String getprop(String name, String defaultValue)
    {
        ProcessBuilder pb = new ProcessBuilder("/system/bin/getprop", name);
        pb.redirectErrorStream(true);

        Process p = null;
        InputStream is = null;
        try
        {
            p = pb.start();
            is = p.getInputStream();
            Scanner scan = new Scanner(is);
            scan.useDelimiter("\n");
            String prop = scan.next();
            if (prop.length() == 0)
            {
                return defaultValue;
            }
            return prop;
        } catch (NoSuchElementException e)
        {
            return defaultValue;
        } catch (Exception e)
        {
            e.printStackTrace();
        } finally
        {
            if (is != null)
            {
                try
                {
                    is.close();
                } catch (Exception e)
                {
                }
            }
        }
        return defaultValue;
    }

    public static void removeConfigFiles(Context context)
    {
        Resources resources = context.getResources();
        String filePath =
                Environment.getExternalStorageDirectory() + resources.getString(R.string.updaterFolder) + resources.getString(R.string.configFilename);

        removeFile(filePath + resources.getString(R.string.config_zip));
        removeZipContents(context);
    }

    public static void removeZipContents(Context context)
    {
        Resources resources = context.getResources();
        String filePath =
                Environment.getExternalStorageDirectory() + resources.getString(R.string.updaterFolder) + resources.getString(R.string.configFilename);

        removeFile(filePath + resources.getString(R.string.config_xml));
        removeFile(filePath + resources.getString(R.string.config_sig));
    }

    private static void removeFile(String filePath)
    {
        File file = new File(filePath);
        if (file.exists())
        {
            file.delete();
        }
    }

}
